"""
Wrapper for subrocess.Popen with interactive input support.
"""
import select
import subprocess
import sys
import threading
from subprocess import Popen
from typing import Sequence, TextIO

from awscliv2.exceptions import ExecutableNotFoundError, SubprocessError


class InteractiveProcess:
    """
    Wrapper for subrocess.Popen with interactive input support.
    """

    read_timeout = 0.2
    default_stdout: TextIO = sys.stdout
    default_stdin: TextIO = sys.stdin

    def __init__(self, command: Sequence[str]) -> None:
        self.command = list(command)
        self.finished = True

    def writeall(self, process: Popen, stdout: TextIO) -> None:
        while True:
            output_data = process.stdout.read(1)
            if not output_data:
                break
            stdout.write(output_data.decode("utf-8"))
            stdout.flush()

    def readall(self, process: Popen, stdin: TextIO) -> None:
        while True:
            if self.finished:
                break

            rlist = select.select([stdin], [], [], self.read_timeout)[0]
            if not rlist:
                continue

            input_data = stdin.readline()
            if not input_data:
                break
            process.stdin.write(input_data.encode())
            process.stdin.flush()

    def run(self, stdin: TextIO = default_stdin, stdout: TextIO = default_stdout) -> int:
        self.finished = False
        try:
            process = Popen(
                self.command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
        except FileNotFoundError:
            raise ExecutableNotFoundError(self.command[0])

        writer = threading.Thread(target=self.writeall, args=(process, stdout))
        reader = threading.Thread(target=self.readall, args=(process, stdin))
        reader.start()
        writer.start()
        try:
            process.wait()
        except KeyboardInterrupt:
            raise SubprocessError("Keyboard interrupt")
        finally:
            self.finished = True
            reader.join()
            writer.join()

        return process.returncode



"""
AWS CLI v2 installers.
"""
import os
import platform
import shutil
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory
from urllib.request import urlopen
from zipfile import ZipFile

from awscliv2.constants import LINUX_ARM_URL, LINUX_X86_64_URL, MACOS_URL
from awscliv2.exceptions import InstallError
from awscliv2.interactive_process import InteractiveProcess
from awscliv2.logger import get_logger


def download_file(url: str, target: Path) -> None:
    """
    Download file from `url` to `target` path.
    """
    get_logger().info(f"Downloading package from {url} to {target}")
    with urlopen(url) as f:
        target.write_bytes(f.read())


def install_macos() -> None:
    """
    Install AWS CLI v2 for MacOS.
    """
    logger = get_logger()
    logger.info("Installing AWS CLI v2 for MacOS")
    install_path = Path.home() / "aws-cli"
    if install_path.exists():
        logger.info(f"Removing {install_path}")
        shutil.rmtree(install_path)

    with NamedTemporaryFile("w+", suffix=".pkg") as f_obj:
        package_path = Path(f_obj.name)
        download_file(MACOS_URL, package_path)
        logger.info(f"Installing {package_path} to {install_path}")
        InteractiveProcess(
            [
                "installer",
                "-pkg",
                package_path.as_posix(),
                "-target",
                install_path.as_posix(),
                "-applyChoiceChangesXML",
                "choices.xml",
            ]
        )
    logger.info("Now awsv2 will use this installed version")
    logger.info("Running now to check installation: awsv2 --version")


def install_linux_x86_64() -> None:
    """
    Install AWS CLI v2 for Linux x86_64.
    """
    install_linux(LINUX_X86_64_URL)


def install_linux_arm() -> None:
    """
    Install AWS CLI v2 for Linux ARM.
    """
    install_linux(LINUX_ARM_URL)


def install_linux(url) -> None:
    """
    Install AWS CLI v2 for Linux from `url`.
    """
    logger = get_logger()
    logger.info("Installing AWS CLI v2 for Linux")
    temp_dir = Path(TemporaryDirectory().name)
    install_path = Path.home() / ".awscliv2"
    if install_path.exists():
        logger.info(f"Removing {install_path}")
        shutil.rmtree(install_path)

    bin_path = install_path / "binaries"
    with NamedTemporaryFile("w+", suffix=".zip") as f_obj:
        package_path = Path(f_obj.name)
        download_file(url, package_path)
        logger.info(f"Extracting {package_path} to to {temp_dir}")
        with ZipFile(package_path, "r") as zip_obj:
            zip_obj.extractall(
                temp_dir,
            )

    installer_path = temp_dir / "aws" / "install"
    os.chmod(installer_path, 0o744)
    os.chmod(temp_dir / "aws" / "dist" / "aws", 0o744)
    logger.info(f"Installing {installer_path} to {install_path}")
    output = StringIO()
    return_code = InteractiveProcess(
        [
            installer_path.as_posix(),
            "-i",
            install_path.as_posix(),
            "-b",
            bin_path.as_posix(),
        ]
    ).run(stdout=output)
    if return_code:
        raise InstallError(f"Installation failed: {output.getvalue()}")
    shutil.rmtree(temp_dir)
    logger.info("Now awsv2 will use this installed version")
    logger.info("Running now to check installation: awsv2 --version")


def install_multiplatform():
    """
    Install AWS CLI v2 for Linux ar MacOS.
    """
    os_platform = platform.system()
    arch = platform.architecture()[0]

    if os_platform == "Darwin":
        return install_macos()
    if os_platform == "Linux" and arch == "64bit":
        return install_linux_x86_64()
    if os_platform == "Linux" and arch == "arm":
        return install_linux_arm()

    raise InstallError(f"{os_platform} {arch} is not supported, use docker version")



IMAGE_NAME = "amazon/aws-cli"
DOCKER_PATH = "docker"
PROG_NAME = "awscliv2"

LINUX_X86_64_URL = "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip"
LINUX_ARM_URL = "https://awscli.amazonaws.com/awscli-exe-linux-aarch64.zip"
MACOS_URL = "https://awscli.amazonaws.com/AWSCLIV2.pkg"



"""
Main entrypoint for CLI.
"""
import json
import sys
from configparser import ConfigParser
from io import StringIO
from pathlib import Path
from typing import List, Optional, Sequence, TextIO

from awscliv2.cli_parser import parse_args
from awscliv2.constants import DOCKER_PATH, IMAGE_NAME
from awscliv2.exceptions import AWSCLIError, ExecutableNotFoundError, SubprocessError
from awscliv2.installers import install_multiplatform
from awscliv2.interactive_process import InteractiveProcess
from awscliv2.logger import get_logger


def run_subprocess(cmd: Sequence[str], stdout: TextIO = sys.stdout) -> int:
    """
    Run interactive subprocess.
    """
    process = InteractiveProcess(cmd)
    try:
        return_code = process.run(stdout=stdout)
    except SubprocessError as e:
        raise AWSCLIError(str(e))

    return return_code


def get_awscli_v2_cmd() -> List[str]:
    local_paths = [
        Path.home() / ".awscliv2" / "binaries" / "aws",
        Path.home() / "aws-cli" / "aws",
    ]
    for local_path in local_paths:
        if local_path.exists():
            return [local_path.as_posix()]

    return [
        DOCKER_PATH,
        "run",
        "-i",
        "--rm",
        "-v",
        f"{Path.home().as_posix()}/.aws:/root/.aws",
        "-v",
        f"{Path.cwd().as_posix()}:/aws",
        IMAGE_NAME,
    ]


def run_awscli_v2(args: Sequence[str], stdout: TextIO = sys.stdout) -> int:
    """
    Run dockerized AWS CLI.
    """
    cmd = [
        *get_awscli_v2_cmd(),
        *args,
    ]
    try:
        return run_subprocess(cmd, stdout=stdout)
    except ExecutableNotFoundError:
        raise AWSCLIError(f"Executable not found: {cmd[0]}")


def run_assume_role(profile_name: str, source_profile: str, role_arn: str) -> None:
    aws_path = Path.home() / ".aws"
    if not aws_path.exists():
        aws_path.mkdir(parents=True, exist_ok=True)
    credentials_path = aws_path / "credentials"
    if not credentials_path.exists():
        credentials_path.write_text("")

    stdout = StringIO()
    return_code = run_awscli_v2(
        [
            "--profile",
            source_profile,
            "sts",
            "assume-role",
            "--role-arn",
            role_arn,
            "--role-session-name",
            f"{profile_name}-{source_profile}",
        ],
        stdout=stdout,
    )
    if return_code:
        raise AWSCLIError(stdout.getvalue().strip())

    credentials_json = stdout.getvalue()
    credentials_data = json.loads(credentials_json)
    set_credentials(
        profile_name=profile_name,
        aws_access_key_id=credentials_data["Credentials"]["AccessKeyId"],
        aws_secret_access_key=credentials_data["Credentials"]["SecretAccessKey"],
        aws_session_token=credentials_data["Credentials"]["SessionToken"],
    )


def set_credentials(
    profile_name: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    aws_session_token: Optional[str] = None,
):
    """
    Add or update credentials in `~/.aws/credentials`
    """
    aws_path = Path.home() / ".aws"
    if not aws_path.exists():
        aws_path.mkdir(parents=True, exist_ok=True)
    credentials_path = aws_path / "credentials"
    if not credentials_path.exists():
        credentials_path.write_text("")

    config = ConfigParser()
    config.read(credentials_path)
    credentials = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
    }
    if aws_session_token:
        credentials["aws_session_token"] = aws_session_token

    config[profile_name] = credentials

    output = StringIO()
    config.write(output)
    credentials_path.write_text(output.getvalue())
    get_logger().info(f"Successfully added {profile_name} profile to {credentials_path.as_posix()}")


def main(args: Sequence[str]) -> None:
    """
    Main program entrypoint.
    """
    namespace = parse_args(args)

    if namespace.update:
        install_multiplatform()
        run_awscli_v2(["--version"])
        sys.exit(0)

    if namespace.install:
        install_multiplatform()
        sys.exit(run_awscli_v2(["--version"]))

    if namespace.version:
        version_path = Path(__file__).parent / "version.txt"
        version = version_path.read_text().strip() if version_path.exists() else "0.0.0"
        print(version)
        cmd = " ".join(get_awscli_v2_cmd())
        print(f"AWS CLI v2 command: {cmd}")
        sys.exit(run_awscli_v2(["--version"]))

    if namespace.assume_role:
        try:
            return run_assume_role(*namespace.assume_role[:3])
        except TypeError:
            raise AWSCLIError("Use --assume-role <name> <source_profile> <role_arn>")

    if namespace.configure:
        try:
            return set_credentials(*namespace.configure[:4])
        except TypeError:
            raise AWSCLIError(
                "Use --configure <profile_name> <access_key> <secret_key> [<session_token>]"
            )

    if not namespace.other:
        raise AWSCLIError("No command provided")

    sys.exit(run_awscli_v2(namespace.other))


def main_cli() -> None:
    """
    Main entrypoint for CLI.
    """
    try:
        main(sys.argv[1:])
    except AWSCLIError as e:
        message = str(e)
        if message:
            logger = get_logger()
            logger.error(message)
        sys.exit(e.returncode)



import logging


def get_logger(level: int = logging.DEBUG) -> logging.Logger:
    logger = logging.getLogger("awscliv2")
    if logger.handlers:
        return logger

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    stream_handler.setLevel(level)
    logger.addHandler(stream_handler)
    logger.setLevel(level)

    return logger



"""
All exceptions for programmatical usage.
"""


class AWSCLIError(BaseException):
    """
    Main error for awscliv2.
    """

    def __init__(self, msg: str = "", returncode: int = 1) -> None:
        self.msg = msg
        self.returncode = returncode

    def __str__(self) -> str:
        return self.msg


class InstallError(AWSCLIError):
    """
    AWS CLi v2 installer error.
    """


class SubprocessError(BaseException):
    """
    Subprocess interrupted error.
    """


class ExecutableNotFoundError(BaseException):
    """
    Subprocess cannot find an executable error.
    """



import argparse
from typing import Sequence

from awscliv2.constants import PROG_NAME


def parse_args(args: Sequence[str]) -> argparse.Namespace:
    """
    Parse CLI arguments.
    """
    parser = argparse.ArgumentParser(PROG_NAME, description="Can be used as a regular AWS CLI v2")
    parser.add_argument(
        "--configure", nargs="+", help="Configure profile: <name> <access_key> <secret_key>"
    )
    parser.add_argument(
        "--assume-role",
        nargs="+",
        help="Configure assume role profile: <name> <source_profile> <role_arn>",
    )
    parser.add_argument("-V", "--version", action="store_true", help="Show version")
    parser.add_argument(
        "-U", "--update", action="store_true", help="Install AWS CLI v2 (deprecated)"
    )
    parser.add_argument("-i", "--install", action="store_true", help="Install AWS CLI v2")
    namespace, other = parser.parse_known_args(args)
    namespace.other = other
    return namespace
